import unittest
import mock

from twitcher.registry import ServiceRegistry


class ServiceRegistryTestCase(unittest.TestCase):
    def setUp(self):
        self.service = dict(name="loving_flamingo", url="http://somewhere.over.the/ocean", type="wps")

    def test_get_service(self):
        collection_mock = mock.Mock(spec=["find_one"])
        collection_mock.find_one.return_value = self.service
        
        registry = ServiceRegistry(collection=collection_mock)
        service = registry.get_service(name=self.service['name'])

        collection_mock.find_one.assert_called_with({"name": self.service['name']})
        assert isinstance(service, dict)

    def test_register_service_default(self):
        collection_mock = mock.Mock(spec=["insert_one", "find_one"])
        collection_mock.find_one.return_value = None
        
        store = ServiceRegistry(collection=collection_mock)
        store.register_service(url=self.service['url'], name=self.service['name'])

        collection_mock.insert_one.assert_called_with(self.service)

    def test_register_service_with_special_name(self):
        collection_mock = mock.Mock(spec=["insert_one", "find_one"])
        collection_mock.find_one.return_value = None
        
        store = ServiceRegistry(collection=collection_mock)
        store.register_service(url="http://wonderload", name="A special Name")

        collection_mock.insert_one.assert_called_with({'url': 'http://wonderload', 'type': 'wps', 'name': 'a_special_name'})



from pyramid.config import Configurator

import logging
logger = logging.getLogger(__name__)

def main(global_config, **settings):
    """
    This function returns a Pyramid WSGI application.
    """
    config = Configurator(settings=settings)

    # beaker session
    config.include('pyramid_beaker')

    # include twitcher components
    config.include('twitcher.frontpage')
    config.include('twitcher.rpcinterface')
    config.include('twitcher.owsproxy')
    config.include('twitcher.wps')
    
    # tweens/middleware
    # TODO: maybe add tween for exception handling or use unknown_failure view
    config.include('twitcher.tweens')
   
    config.scan()

    return config.make_wsgi_app()




from nose.tools import assert_equals, assert_raises

from twitcher import utils

def test_baseurl():
    assert_equals(utils.baseurl('http://localhost:8094/wps'), 'http://localhost:8094/wps')
    assert_equals(utils.baseurl('http://localhost:8094/wps?service=wps&request=getcapabilities'), 'http://localhost:8094/wps')
    assert_equals(utils.baseurl('https://localhost:8094/wps?service=wps&request=getcapabilities'), 'https://localhost:8094/wps')
    with assert_raises(ValueError) as e:
        utils.baseurl('ftp://localhost:8094/wps')


def test_path_elements():
    assert_equals(utils.path_elements('/ows/proxy/lovely_bird'), ['ows', 'proxy', 'lovely_bird'])
    assert_equals(utils.path_elements('/ows/proxy/lovely_bird/'), ['ows', 'proxy', 'lovely_bird'])
    assert_equals(utils.path_elements('/ows/proxy/lovely_bird/ '), ['ows', 'proxy', 'lovely_bird'])


def test_lxml_strip_ns():
    import lxml.etree
    wpsxml = """<wps100:Execute xmlns:wps100="http://www.opengis.net/wps/1.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" service="WPS" version="1.0.0" xsi:schemaLocation="http://www.opengis.net/wps/1.0.0 http://schemas.opengis.net/wps/1.0.0/wpsExecute_request.xsd"/>"""

    doc = lxml.etree.fromstring(wpsxml)
    assert_equals(doc.tag, '{http://www.opengis.net/wps/1.0.0}Execute')
    utils.lxml_strip_ns(doc)
    assert_equals(doc.tag, 'Execute')



"""
pywps wrapper
"""
import os

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.settings import asbool, aslist

import pywps
from pywps.Exceptions import WPSException
from twitcher.owsexceptions import OWSNoApplicableCode

import logging
logger = logging.getLogger(__name__)

DEFAULT_KEYS =  ['PYWPS_CFG', 'PYWPS_PROCESSES', 'PYWPS_TEMPLATES']

def _wps_environ_keys(request):
    settings = request.registry.settings
    if 'twitcher.wps_environ_keys' in settings:
        keys = aslist(settings['twitcher.wps_environ_keys'])
    else:
        keys = DEFAULT_KEYS
    return keys


def _wps_cfg(request):
    settings = request.registry.settings
    return settings.get('twitcher.wps_cfg')

    
def pywps_view(request):
    """
    * TODO: add xml response renderer
    * TODO: fix exceptions ... use OWSException (raise ...)
    """
    response = request.response
    response.status = "200 OK"
    response.content_type = "text/xml"

    inputQuery = None
    os.environ["REQUEST_METHOD"] = request.method
    if request.method == "GET":
        inputQuery = request.query_string
    elif request.method == "POST":
        inputQuery = request.body_file_raw
    else:
        return HTTPBadRequest()

    if not inputQuery:
        return OWSNoApplicableCode("No query string found.")

    if request.wps_cfg:
        os.environ['PYWPS_CFG'] = request.wps_cfg

    # set the environ for wps from request environ
    for key in request.wps_environ_keys:
        if key in request.environ:
            os.environ[key] = request.environ[key]

    # create the WPS object
    try:
        wps = pywps.Pywps(os.environ["REQUEST_METHOD"], os.environ.get("PYWPS_CFG"))
        if wps.parseRequest(inputQuery):
            pywps.debug(wps.inputs)
            wps.performRequest(processes=os.environ.get("PYWPS_PROCESSES"))
            response_headers = [('Content-type', wps.request.contentType)]
            return wps.response
    except WPSException,e:
        return str(e)
    except Exception, e:
        return OWSNoApplicableCode(e.message)

def includeme(config):
    settings = config.registry.settings

    if asbool(settings.get('twitcher.wps', True)):
        logger.info('Add twitcher wps application')

        config.add_route('wps', '/ows/wps')
        config.add_route('wps_secured', '/ows/wps/{access_token}')
        config.add_view(pywps_view, route_name='wps', renderer='string')
        config.add_view(pywps_view, route_name='wps_secured', renderer='string')
        config.add_request_method(_wps_environ_keys, 'wps_environ_keys', reify=True)
        config.add_request_method(_wps_cfg, 'wps_cfg', reify=True)






from pyramid import testing

from twitcher.tokens import tokenstore_factory, tokengenerator_factory

def setup_with_db():
    settings = {'mongodb.host':'127.0.0.1', 'mongodb.port':'27027', 'mongodb.db_name': 'twitcher_test'}
    config = testing.setUp(settings=settings)
    return config

def setup_tokenstore(config):
    store = tokenstore_factory(config.registry)
    generator = tokengenerator_factory(config.registry)
    store.clean_tokens()
    access_token = generator.create_access_token()
    store.save_token(access_token)
    return access_token.token

      



from nose.tools import ok_, assert_raises
import unittest
import mock

from pyramid.testing import DummyRequest

from twitcher.tokens import AccessToken, expires_at
from twitcher.owssecurity import OWSSecurity
from twitcher.owsexceptions import OWSAccessForbidden


class OWSSecurityTestCase(unittest.TestCase):
    def setUp(self):
        self.access_token = AccessToken(token="cdefg", expires_at=expires_at(hours=1))
        
        store_mock = mock.Mock(spec=["fetch_by_token"])
        store_mock.fetch_by_token.return_value = self.access_token
        self.security = OWSSecurity(tokenstore=store_mock)

    def test_get_token_by_param(self):
        params = dict(request="Execute", service="WPS", access_token="abcdef")
        request = DummyRequest(params=params)
        token = self.security.get_token_param(request)
        ok_(token == "abcdef")


    def test_get_token_by_path(self):
        params = dict(request="Execute", service="WPS")
        request = DummyRequest(params=params, path="/ows/emu/12345")
        token = self.security.get_token_param(request)
        ok_(token == "12345")


    def test_get_token_by_header(self):
        params = dict(request="Execute", service="WPS")
        headers = {'Access-Token': '54321'}
        request = DummyRequest(params=params, headers=headers)
        token = self.security.get_token_param(request)
        ok_(token == "54321")


    def test_check_request(self):
        params = dict(request="Execute", service="WPS", version="1.0.0",  access_token="cdefg")
        request = DummyRequest(params=params, path='/ows/emu')
        self.security.check_request(request)

        
    def test_check_request_invalid(self):
        store_mock = mock.Mock(spec=["fetch_by_token"])
        store_mock.fetch_by_token.return_value = None
        security = OWSSecurity(tokenstore=store_mock)
        
        params = dict(request="Execute", service="WPS", version="1.0.0", access_token="xyz")
        request = DummyRequest(params=params, path='/ows/emu')
        with assert_raises(OWSAccessForbidden):
            security.check_request(request)


    def test_check_request_allowed_caps(self):
        store_mock = mock.Mock(spec=["fetch_by_token"])
        store_mock.fetch_by_token.return_value = None
        security = OWSSecurity(tokenstore=store_mock)
        
        params = dict(request="GetCapabilities", service="WPS", version="1.0.0")
        request = DummyRequest(params=params, path='/ows/emu')
        security.check_request(request)

    def test_check_request_allowed_describeprocess(self):
        store_mock = mock.Mock(spec=["fetch_by_token"])
        store_mock.fetch_by_token.return_value = None
        security = OWSSecurity(tokenstore=store_mock)
        
        params = dict(request="DescribeProcess", service="WPS", version="1.0.0")
        request = DummyRequest(params=params, path='/ows/emu')
        security.check_request(request)


    






"""
Classes to manage access tokens used in the security middleware.

The implementation is based on `python-oauth2 <http://python-oauth2.readthedocs.io/en/latest/>`_

See access token examples:

* https://www.mapbox.com/developers/api/
* http://python-oauth2.readthedocs.io/en/latest/store.html
"""

import uuid
import time

from twitcher.utils import now_secs
from twitcher.exceptions import AccessTokenNotFound
from twitcher.db import mongodb

import logging
logger = logging.getLogger(__name__)


def tokenstore_factory(registry):
    db = mongodb(registry)
    return MongodbAccessTokenStore(db.tokens)


def tokengenerator_factory(registry):
    return UuidGenerator()


def expires_at(hours=1):
    return now_secs() + hours * 3600 

class AccessTokenStore(object):

    def save_token(self, access_token):
        """
        Stores an access token with additional data.
        """
        raise NotImplementedError

          
    def delete_token(self, token):
        """
        Deletes an access token from the store using its token string to identify it.
        This invalidates both the access token and the token.

        :param token: A string containing the token.
        :return: None.
        """
        raise NotImplementedError

    
    def fetch_by_token(self, token):
        """
        Fetches an access token from the store using its token string to
        identify it.

        :param token: A string containing the token.
        :return: An instance of :class:`twitcher.tokens.AccessToken`.
        """
        raise NotImplementedError


    def clean_tokens(self):
        """
        Removes all tokens from database.
        """
        raise NotImplementedError



class MongodbAccessTokenStore(AccessTokenStore):
    def __init__(self, collection):
        self.collection = collection

    def save_token(self, access_token):
        self.collection.insert_one(access_token)

          
    def delete_token(self, token):
        self.collection.delete_one({'token': token})

    
    def fetch_by_token(self, token):
        token = self.collection.find_one({'token': token})
        if not token:
            raise AccessTokenNotFound
        return AccessToken(token)


    def clean_tokens(self):
        self.collection.drop()


class AccessTokenGenerator(object):
    """
    Base class for access token generators.
    """
    def create_access_token(self, valid_in_hours=1, user_environ=None):
        """
        Creates an access token.

        TODO: check valid in hours
        TODO: maybe specify how often a token can be used
        """
        access_token = AccessToken(
            token = self.generate(),
            expires_at = expires_at(hours=valid_in_hours),
            user_environ = user_environ)
        return access_token
    
    def generate(self):
        raise NotImplementedError

    
class UuidGenerator(AccessTokenGenerator):
    """
    Generate a token using uuid4.
    """
    def generate(self):
        """
        :return: A new token
        """
        return uuid.uuid4().get_hex()

    
class AccessToken(dict):
    """
    Dictionary that contains access token. It always has ``'token'`` key.
    """
    
    def __init__(self, *args, **kwargs):
        super(AccessToken, self).__init__(*args, **kwargs)
        if 'token' not in self:
            raise TypeError("'token' is required")

        self.expires_at = int( self.get("expires_at", 0) )

            
    @property
    def token(self):
        """Access token string."""
        return self['token']


    @property
    def expires_in(self):
        """
        Returns the time until the token expires.
        :return: The remaining time until expiration in seconds or 0 if the
                 token has expired.
        """
        time_left = self.expires_at - now_secs()

        if time_left > 0:
            return time_left
        return 0

    
    def is_expired(self):
        """
        Determines if the token has expired.
        :return: `True` if the token has expired. Otherwise `False`.
        """
        if self.expires_at is None:
            return True

        if self.expires_in > 0:
            return False

        return True

    
    @property
    def user_environ(self):
        environ = self.get('user_environ') or {}
        return environ

    
    def __str__(self):
        return self.token

    
    def __repr__(self):
        cls = type(self)
        repr_ = dict.__repr__(self)
        return '{0}.{1}({2})'.format(cls.__module__, cls.__name__, repr_)



    







"""
OWSExceptions are based on pyramid.httpexceptions.

See also: https://github.com/geopython/pywps/blob/master/pywps/exceptions.py
"""


from string import Template

from zope.interface import implementer

from webob import html_escape as _html_escape

from pyramid.interfaces import IExceptionResponse
from pyramid.response import Response

@implementer(IExceptionResponse)
class OWSException(Response, Exception):

    code = 'NoApplicableCode'
    value = None
    locator = 'NoApplicableCode'
    explanation = 'Unknown Error'

    page_template = Template('''\
<?xml version="1.0" encoding="utf-8"?>
<ExceptionReport version="1.0.0" xmlns="http://www.opengis.net/ows/1.1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.opengis.net/ows/1.1 http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd">
    <Exception exceptionCode="${code}" locator="${locator}">
        <ExceptionText>${message}</ExceptionText>
    </Exception>
</ExceptionReport>''')


    def __init__(self, detail=None, value=None, **kw):
        Response.__init__(self, status='200 OK', **kw)
        Exception.__init__(self, detail)
        self.message = detail or self.explanation
        if value:
            self.locator = value

    def __str__(self):
        return self.message

    def prepare(self, environ):
        if not self.body:
            self.content_type = 'text/xml'
            args = {
                'code': _html_escape(self.code),
                'locator': _html_escape(self.locator),
                'message': _html_escape(self.message or ''),
                }
            page = self.page_template.substitute(args)
            page = page.encode(self.charset)
            self.app_iter = [page]
            self.body = page

    @property
    def wsgi_response(self):
        # bw compat only
        return self

    exception = wsgi_response # bw compat only

    def __call__(self, environ, start_response):
        # differences from webob.exc.WSGIHTTPException
        #
        # - does not try to deal with HEAD requests
        #
        # - does not manufacture a new response object when generating
        #   the default response
        #
        self.prepare(environ)
        return Response.__call__(self, environ, start_response)

class OWSAccessForbidden(OWSException):
    locator = "AccessForbidden"
    explanation = "Access to this service is forbidden"

    
class OWSNoApplicableCode(OWSException):
    pass


class OWSMissingParameterValue(OWSException):
    """MissingParameterValue WPS Exception"""
    code = "MissingParameterValue"
    locator = ""
    explanation = "Parameter value is missing"
    
class OWSInvalidParameterValue(OWSException):
    """InvalidParameterValue WPS Exception"""
    code = "InvalidParameterValue"
    locator = ""
    explanation = "Parameter value is invalid"
    



"""
Based on unitests in https://github.com/wndhydrnt/python-oauth2/tree/master/oauth2/test
"""

from nose.tools import ok_, assert_raises
import unittest
import mock

from twitcher.tokens import AccessToken, UuidGenerator, MongodbAccessTokenStore, expires_at


class MongodbAccessTokenStoreTestCase(unittest.TestCase):
    def setUp(self):
        self.access_token = AccessToken(token="abcdef", expires_at=expires_at(hours=1))

    def test_fetch_by_token(self):
        collection_mock = mock.Mock(spec=["find_one"])
        collection_mock.find_one.return_value = self.access_token
        
        store = MongodbAccessTokenStore(collection=collection_mock)
        access_token = store.fetch_by_token(token=self.access_token.token)

        collection_mock.find_one.assert_called_with({"token": self.access_token.token})
        ok_(isinstance(access_token, AccessToken))

    
    def test_save_token(self):
        collection_mock = mock.Mock(spec=["insert_one"])
        
        store = MongodbAccessTokenStore(collection=collection_mock)
        store.save_token(self.access_token)

        collection_mock.insert_one.assert_called_with(self.access_token)

        
class UuidGeneratorTestCase(unittest.TestCase):
    def setUp(self):
        self.generator = UuidGenerator()
    
    def test_generate(self):
        token = self.generator.generate()
        ok_(len(token) == 32)

    def test_create_access_token_default(self):
        access_token = self.generator.create_access_token()
        ok_(len( access_token.token ) == 32)
        ok_(access_token.expires_in <= 3600)

    def test_create_access_non_default_hours(self):
        access_token = self.generator.create_access_token(valid_in_hours=2)
        ok_(len( access_token.token ) == 32)
        ok_(access_token.expires_in <= 3600 * 2)
        

class AccessTokenTestCase(unittest.TestCase):

    def test_access_token(self):
        access_token = AccessToken(token='abcdef', expires_at=expires_at(hours=1))
        ok_(access_token.expires_in <= 3600)
        ok_(access_token.is_expired() == False)

    def test_missing_token(self):
        with assert_raises(TypeError) as e:
            AccessToken()

    def test_invalid_access_token(self):
        access_token = AccessToken(token='abcdef', expires_at=expires_at(hours=-1))
        ok_(access_token.expires_in == 0)
        ok_(access_token.is_expired() == True)


    def test_access_token_with_user_environ(self):
        access_token = AccessToken(token='12345', expires_at=expires_at(hours=1),
                            user_environ={'data_token': 'bfghk'})
        ok_(access_token.user_environ == {'data_token': 'bfghk'})



import pymongo

from twitcher.utils import namesgenerator, baseurl
from twitcher.db import mongodb

import logging
logger = logging.getLogger(__name__)


def proxy_url(request, service_name):
    """
    Shortcut method to return route url to service name.
    """
    return request.route_url('owsproxy', service_name=service_name)


def update_with_proxy_url(request, services):
    for service in services:
        service['proxy_url'] = proxy_url(request, service['name'])
    return services

def service_name_of_proxy_url(proxy_url):
    from urlparse import urlparse
    parsed_url = urlparse(proxy_url)
    service_name = None
    if parsed_url.path.startswith("/ows/proxy"):
        service_name = parsed_url.path.strip('/').split('/')[2]
    return service_name


def service_registry_factory(registry):
    db = mongodb(registry)
    return ServiceRegistry(collection=db.services)


class ServiceRegistry(object):
    """
    Registry for OWS services. Uses mongodb to store service url and attributes. 
    """
    
    def __init__(self, collection):
        self.collection = collection

    def register_service(self, url, name=None, service_type='wps'):
        """
        Adds OWS service with given name to registry database.
        """
        
        service_url = baseurl(url)
        # check if service is already registered
        service = self.collection.find_one({'url': service_url})
        if service is None:
            name = namesgenerator.get_sane_name(name)
            if name is None:
                name = namesgenerator.get_random_name()
                if not self.collection.find_one({'name': name}) is None:
                    name = namesgenerator.get_random_name(retry=True)
            service = dict(url=service_url, name=name, type=service_type)
            if self.collection.find_one({'name': name}):
                logging.info("update registered service %s." % (name))
                self.collection.update_one({'name': name}, {'$set': service})
            else:
                self.collection.insert_one(service)
            service = self.collection.find_one({'name': service['name']})
        return service


    def unregister_service(self, name):
        """
        Removes service from registry database.
        """
        self.collection.delete_one({'name': name})


    def list_services(self):
        """
        Lists all servcies in registry database.
        """
        my_services = []
        for service in self.collection.find().sort('name', pymongo.ASCENDING):
            my_services.append({
                'name': service['name'],
                'type': service['type'],
                'url': service['url']})
        return my_services


    def get_service(self, name):
        """
        Get service for given ``name`` from registry database.
        """
        service = self.collection.find_one({'name': name})
        if service is None:
            raise ValueError('service not found')
        if not 'url' in service:
            raise ValueError('service has no url')
        return dict(url=service.get('url'), name=name)


    def get_service_by_url(self, url):
        """
        Get service for given ``url`` from registry database.
        """
        service = self.collection.find_one({'url': url})
        if service is None:
            raise ValueError('service not found')
        return dict(name=service.get('name'), url=url)
    
    def clear_services(self):
        """
        Removes all OWS services from registry database.
        """
        self.collection.drop()




    
    








"""
The owsproxy is based on `papyrus_ogcproxy <https://github.com/elemoine/papyrus_ogcproxy>`_
"""

import urllib
import requests

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPBadRequest, HTTPBadGateway, HTTPNotAcceptable, HTTPForbidden
from twitcher.owsexceptions import OWSNoApplicableCode, OWSAccessForbidden
from pyramid.response import Response
from pyramid.settings import asbool

from twitcher.registry import service_registry_factory, proxy_url

import logging
logger = logging.getLogger(__name__)


allowed_content_types = (
    "application/xml", "text/xml",
    "application/vnd.ogc.se_xml",           # OGC Service Exception
    "application/vnd.ogc.se+xml",           # OGC Service Exception
    #"application/vnd.ogc.success+xml",      # OGC Success (SLD Put)
    #"application/vnd.ogc.wms_xml",          # WMS Capabilities
    #"application/vnd.ogc.gml",              # GML
    #"application/vnd.ogc.sld+xml",          # SLD
    #"application/vnd.google-earth.kml+xml", # KML
    )

          
def _send_request(request, service):
    # TODO: fix way to build url
    logger.debug('params = %s', request.params)
    url = service['url'] + '?' + urllib.urlencode(request.params)

    # forward request to target (without Host Header)
    h = dict(request.headers)
    h.pop("Host", h)
    resp = None
    try:
        resp = requests.request(method=request.method.upper(), url=url, data=request.body, headers=h)
    except Exception, e:
        return HTTPBadGateway(e.message)

    if resp.ok == False:
        return HTTPBadGateway(resp.reason)

    # check for allowed content types
    ct = None
    if "Content-Type" in resp.headers:
        ct = resp.headers["Content-Type"]
        if not ct.split(";")[0] in allowed_content_types:
            return OWSAccessForbidden()
    else:
        return HTTPNotAcceptable()

    content = None
    try:
        content = resp.content.decode('utf-8', 'ignore')
        # replace urls in xml content
        if ct in ['text/xml', 'application/xml']:
            content = content.replace(service['url'], proxy_url(request, service['name']))
    except:
        return HTTPNotAcceptable("Could not decode content.")

    return Response(content, status=resp.status_code, headers={"Content-Type": ct})

def owsproxy_view(request):
    """
    TODO: use ows exceptions
    """
    service_name = request.matchdict.get('service_name')
    if service_name is None:
        return HTTPBadRequest('Parameter service_name is required.')

    try:
        registry = service_registry_factory(request.registry)
        service = registry.get_service(service_name)
    except Exception as err:
        return HTTPBadRequest("Could not find service: %s." % (err.message))
    else:
        return _send_request(request, service)

def includeme(config):
    settings = config.registry.settings

    if asbool(settings.get('twitcher.ows_proxy', True)):
        logger.info('Add OWS proxy')
        
        # include mongodb
        config.include('twitcher.db')
    
        config.add_route('owsproxy', '/ows/proxy/{service_name}')
        config.add_route('owsproxy_secured', '/ows/proxy/{service_name}/{access_token}')
        config.add_view(owsproxy_view, route_name='owsproxy')
        config.add_view(owsproxy_view, route_name='owsproxy_secured')



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()

reqs = [line.strip() for line in open('requirements/deploy.txt')]
test_reqs = [line.strip() for line in open('requirements/tests.txt')]

setup(name='pyramid_twitcher',
      version='0.1.7',
      description='Security Proxy for OGC Services like WPS.',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Development Status :: 4 - Beta",
        ],
      author='Birdhouse Developers',
      author_email='wps-dev@lists.dkrz.de',
      url='https://github.com/bird-house/twitcher.git',
      license='Apache License 2.0',
      keywords='buildout pyramid twitcher birdhouse wps pywps security proxy ows ogc',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='twitcher',
      install_requires=reqs,
      entry_points="""\
      [paste.app_factory]
      main = twitcher:main
      [console_scripts]
      twitcherctl=twitcher.twitcherctl:main
      """,
      )



import functools

from pyramid.view import view_defaults
from pyramid_rpc.xmlrpc import xmlrpc_method
from pyramid.settings import asbool

from twitcher.registry import service_registry_factory, update_with_proxy_url
from twitcher.tokens import tokengenerator_factory
from twitcher.tokens import tokenstore_factory

import logging
logger = logging.getLogger(__name__)


# shortcut for xmlrpc_method
api_xmlrpc = functools.partial(xmlrpc_method, endpoint="api")


@view_defaults(permission='admin')
class RPCInterface(object):
    def __init__(self, request):
        self.request = request
        registry = self.request.registry
        self.tokengenerator = tokengenerator_factory(registry)
        self.tokenstore = tokenstore_factory(registry)
        self.registry = service_registry_factory(registry)

    # token management

    @api_xmlrpc()
    def gentoken(self, valid_in_hours=1, user_environ=None):
        """
        Generates an access token which is valid for ``valid_in_hours``. Stores the optional ``user_environ`` dict with the token.
        """
        access_token = self.tokengenerator.create_access_token(valid_in_hours=valid_in_hours, user_environ=user_environ)
        self.tokenstore.save_token(access_token)
        return access_token.token

    @api_xmlrpc()
    def revoke(self, token):
        """
        Remove token from tokenstore.
        """
        self.tokenstore.delete_token(token)

    
    @api_xmlrpc()
    def clean(self):
        """
        Removes all tokens.
        """
        try:
            self.tokenstore.clean_tokens()
        except:
            logger.exception('clean tokens failed')
            return False
        else:
            return True
            

    # service registry

    @api_xmlrpc()
    def register(self, url, name=None):
        """
        Adds an OWS service with the given ``url`` to the registry.
        """
        service = self.registry.register_service(url=url, name=name)
        return service['name']


    @api_xmlrpc()
    def unregister(self, name):
        """
        Removes OWS service with the given ``name`` from the registry.
        """
        try:
            self.registry.unregister_service(name=name)
        except:
            logger.exception('unregister failed')
            return False
        else:
            return True



    @api_xmlrpc()
    def status(self):
        """
        Lists all registred OWS services.
        """
        try:
            services = self.registry.list_services()
            return update_with_proxy_url(self.request, services)
        except:
            logger.exception('register failed')
            return []


    @api_xmlrpc()
    def purge(self):
        """
        Removes all services from the registry.
        """
        try:
            self.registry.clear_services()
        except:
            logger.exception('clear failed')
            return False
        else:
            return True

def includeme(config):
    """ The callable makes it possible to include rpcinterface
    in a Pyramid application.

    Calling ``config.include(twitcher.rpcinterface)`` will result in this
    callable being called.

    Arguments:

    * ``config``: the ``pyramid.config.Configurator`` object.
    """
    settings = config.registry.settings

    if asbool(settings.get('twitcher.rpcinferface', True)):
        logger.info('Add twitcher rpcinterface')

        # using basic auth
        config.include('twitcher.basicauth')
    
        # pyramid xml-rpc
        # http://docs.pylonsproject.org/projects/pyramid-rpc/en/latest/xmlrpc.html
        config.include('pyramid_rpc.xmlrpc')
        config.include('twitcher.db')
        config.add_xmlrpc_endpoint('api', '/RPC2')




from twitcher.exceptions import AccessTokenNotFound
from twitcher.owsexceptions import OWSAccessForbidden, OWSInvalidParameterValue
from twitcher.utils import path_elements
from twitcher.tokens import tokenstore_factory
from twitcher.owsrequest import OWSRequest


allowed_service_types = ('wps',)
allowed_request_types = ('getcapabilities', 'describeprocess')
protected_path = '/ows/'

def owssecurity_factory(registry):
    return OWSSecurity(tokenstore_factory(registry))


class OWSSecurity(object):

    def __init__(self, tokenstore):
        self.tokenstore = tokenstore

    
    def get_token_param(self, request):
        token = None
        if 'access_token' in request.params:
            token = request.params['access_token']   # in params
        elif 'Access-Token' in request.headers:
            token = request.headers['Access-Token']  # in header
        else:  # in path
            elements = path_elements(request.path)
            if len(elements) > 1: # there is always /ows/
                token = elements[-1]   # last path element
        return token

    
    def check_request(self, request):
        if request.path.startswith(protected_path):
            ows_request = OWSRequest(request)
            if not ows_request.service in allowed_service_types:
                raise OWSInvalidParameterValue(
                    "service %s not supported" % ows_request.service, value="service")
            if not ows_request.request in allowed_request_types:
                try:
                    token = self.get_token_param(request)
                    access_token = self.tokenstore.fetch_by_token(token)
                    if not access_token:
                        raise AccessTokenNotFound()
                    elif access_token.is_expired():
                        raise OWSAccessForbidden("Access token is expired.")
                    # update request with user environ from access token
                    request.environ.update( access_token.user_environ )
                except AccessTokenNotFound:
                    raise OWSAccessForbidden("Access token is required to access this service.")
            
                
        



import getpass
import argcomplete
import argparse
import xmlrpclib
import ssl
from urlparse import urlparse

import logging
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.WARN)
logger = logging.getLogger(__name__)


def _create_https_context(verify=True):
    context = ssl._create_default_https_context()
    if verify == False:
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
    return context

def _create_server(url, verify_ssl=True, username=None, password=None):
    # TODO: disable basicauth when username is not set
    username = username or 'nouser'
    password = password or 'nopass'

    parsed = urlparse(url)
    url = "%s://%s:%s@%s%s" % (parsed.scheme, username, password, parsed.netloc, parsed.path)
    context = _create_https_context(verify=verify_ssl)
    server = xmlrpclib.ServerProxy(url, context=context)
    return server

class TwitcherCtl(object):
    """
    Command line to interact with the xmlrpc interface of the ``twitcher`` service.
    """

    def create_parser(self):
        parser = argparse.ArgumentParser(
            prog="twitcherctl",
            description='twitcherctl -- control twitcher service from the cmd line.',
            )
        parser.add_argument("--debug",
                            help="Enable debug mode.",
                            action="store_true")
        parser.add_argument('-s', '--serverurl',
                            metavar='URL',
                            default='https://localhost:38083',
                            help='URL on which twitcher server is listening (default "https://localhost:38083").')
        parser.add_argument("-u", "--username",
                            help="Username to use for authentication with server.")
        parser.add_argument("-p", "--password",
                            help="Password to use for authentication with server")
        parser.add_argument("-k", "--insecure", # like curl
                            help="Don't validate the server's certificate.",
                            action="store_true")

        # commands
        subparsers = parser.add_subparsers(
            dest='cmd',
            title='command',
            description='List of available commands',
            )

        # token
        # -----
        subparser = subparsers.add_parser('gentoken', help="Generates an access token.")
        subparser.add_argument('-H', '--valid-in-hours', type=int, default=1, help="Set how long the token is valid in hours (default: 1 hour).")
        subparser.add_argument('-e', '--env', nargs='*', default=[], help="Set environment variable (key=value).")
        
        subparser = subparsers.add_parser('revoke', help="Removes given access token.")
        subparser.add_argument('token', help="access token")
        
        subparser = subparsers.add_parser('clean', help="Removes all access tokens.")
        

        # service registry
        # ----------------
        
        # status
        subparser = subparsers.add_parser('status', help="Lists all registered OWS services used by OWS proxy.")

        # purge
        subparser = subparsers.add_parser('purge', help="Removes all OWS services from the registry.")

        # register
        subparser = subparsers.add_parser('register', help="Adds OWS service to the registry to be used by the OWS proxy.")
        subparser.add_argument('url', help="Service url.")
        subparser.add_argument('--name', help="Service name. If not set then a name will be generated.")

        # unregister
        subparser = subparsers.add_parser('unregister', help="Removes OWS service from the registry.")
        subparser.add_argument('name', help="Service name.")

        return parser

    def run(self, args):
        if args.debug:
            logger.setLevel(logging.DEBUG)

        if args.insecure:
            logger.warn('disabled certificate verification!')

        password = args.password
        if args.username:
            #username = raw_input('Username:')
            if not password:
                password = getpass.getpass(prompt='Password:')

        verify_ssl = args.insecure == False
        server = _create_server(
            url=args.serverurl, verify_ssl=verify_ssl,
            username=args.username, password=password)
        result = None
        try:
            if args.cmd == 'status':
                result = server.status()
            elif args.cmd == 'register':
                if args.name:
                    result = server.register(args.url, args.name)
                else:
                    result = server.register(args.url)
            elif args.cmd == 'unregister':
                result = server.unregister(args.name)
            elif args.cmd == 'purge':
                result = server.purge()
            elif args.cmd == 'gentoken':
                user_environ = {k:v for k,v in (x.split('=') for x in args.env) }
                result = server.gentoken(args.valid_in_hours, user_environ)
            elif args.cmd == 'revoke':
                result = server.revoke(args.token)
            elif args.cmd == 'clean':
                result = server.clean()
        except xmlrpclib.Fault as e:
            logger.error("A fault occurred: %s (%d)", e.faultString, e.faultCode)
        except xmlrpclib.ProtocolError as e:
            logger.error("A protocol error occurred. URL: %s, HTTP/HTTPS headers: %s, Error code: %d, Error message: %s",  e.url, e.headers, e.errcode, e.errmsg)
        except xmlrpclib.ResponseError as e:
            logger.error("A response error occured. Maybe service needs authentication with username and password? %s", e.message)
        except Exception as e:
            logger.error('Unknown error occured. Maybe you need to use the "--insecure" option to access the service on HTTPS? %s', e.message)
        else:
            return result

def main():
    logger.setLevel(logging.INFO)

    ctl = TwitcherCtl()
    parser = ctl.create_parser()
    argcomplete.autocomplete(parser)

    args = parser.parse_args()
    return ctl.run(args)

if __name__ == '__main__':
    sys.exit(main())



import time
from datetime import datetime
import pytz
from urlparse import urlparse

import logging
logger = logging.getLogger(__name__)

def now():
    return localize_datetime(datetime.utcnow())


def now_secs():
    """
    Return the current time in seconds since the Epoch.
    """
    return int(time.time())


def localize_datetime(dt, tz_name='UTC'):
    """Provide a timzeone-aware object for a given datetime and timezone name
    """
    tz_aware_dt = dt
    if dt.tzinfo is None:
        utc = pytz.timezone('UTC')
        aware = utc.localize(dt)
        timezone = pytz.timezone(tz_name)
        tz_aware_dt = aware.astimezone(timezone)
    else:
        logger.warn('tzinfo already set')
    return tz_aware_dt

def baseurl(url):
    """
    return baseurl of given url
    """
    parsed_url = urlparse(url)
    if not parsed_url.netloc or not parsed_url.scheme in ("http", "https"):
        raise ValueError('bad url')
    service_url = "%s://%s%s" % (parsed_url.scheme, parsed_url.netloc, parsed_url.path.strip())
    return service_url


def path_elements(path):
    elements = [el.strip() for el in path.split('/')]
    elements = [el for el in elements if len(el) > 0]
    return elements


def lxml_strip_ns(tree):
    for node in tree.iter():
        try:
            has_namespace = node.tag.startswith('{')
        except AttributeError:
            continue  # node.tag is not a string (node is a comment or similar)
        if has_namespace:
            node.tag = node.tag.split('}', 1)[1]


